proc card {cardNo} {
  if {$cardNo==""} {return 0}
  if {$cardNo>=0} {return $cardNo}
  return [expr -(1+$cardNo)]
}

proc cardList {cardList} {
  set ret {}
  foreach c $cardList {lappend ret [card $c]}
  return $ret
}

proc cardFlip {cardList} {
  # flips a list of cards
  set ret {}
  foreach c $cardList {
    lappend ret [expr -$c-1]
  }
  return $ret
}

proc mix {{seed -1}} {
  # mix CARDS
  global CARDS

  if {$seed!=-1} {expr srand($seed)}
  set tmp $CARDS
  set CARDS {}
  set l [llength $tmp]
  for {set i 0} {$i<$l} {incr i} {
    set r [expr round(rand()*($l-$i-1))]
    lappend CARDS [lindex $tmp $r]
    set tmp [lreplace $tmp $r $r]
  }
}

proc cardFlipped {cardList} {
  # return 0, if all cards are open, 1 if all are flipped and -1 if they mix
  set ret -2
  foreach c $cardList {
    if {$c<0} {
      if {$ret==-2} {
        set ret 1
      } elseif {$ret==0} {
        return -1
      }
    } else {
      if {$ret==-2} {
        set ret 0
      } elseif {$ret==1} {
        return -1
      }
    }
  }
  if {$ret==-2} {set ret -1}
  return $ret
}

proc cardPicture {cardList} {
  # return the common picture of the cards 0..3 (Karo..Kreuz); -1 if none
  set ret -2
  foreach c $cardList {
    set col [expr [card $c]%4]
    if {$ret==-2} {
      set ret $col
    } elseif {$ret!=$col} {
      return -1
    }
  }
  if {$ret==-2} {set ret -1}
  return $ret
}

proc cardColor {cardList} {
  # return the common color of the cards 0 (red) or 1 (black); -1 if none
  set ret -2
  foreach c $cardList {
    set col [expr ([card $c]/2)%2]
    if {$ret==-2} {
      set ret $col
    } elseif {$ret!=$col} {
      return -1
    }
  }
  if {$ret==-2} {set ret -1}
  return $ret
}

proc cardValue {cardList} {
  # return the common value of the cards 0..12 (As,2,3,...,10,B,D,K); -1 if none
  set ret -2
  foreach c $cardList {
    set val [expr ([card $c]%52)/4]
    if {$ret==-2} {
      set ret $val
    } elseif {$ret!=$val} {
      return -1
    }
  }
  if {$ret==-2} {set ret -1}
  return $ret
}

proc cardAltColors {cardList} {
  # return whether the cards follow with alternate colors (red black red...)
  set ret -2
  foreach c $cardList {
    set col [expr ([card $c]/2)%2]
    if {$ret==-2} {
      set vgl $col
      set ret 1
    } elseif {$col==$vgl} {
      return 0
    }
    set vgl $col
  }
  if {$ret==-2} {return 0}
  return 1
}

proc cardSequence {cardList} {
  # return iff the cardList is an ascending (1), descending (0) or no (-1)
  # sequence of cards. 1 is checked before 0. A sequence is a loose row
  # like 2,4,5,5,10,...
  set ret -3
  foreach c $cardList {
    set val [cardValue [list $c]]
    if {$ret==-3} {
      set ret -2
    } elseif {$ret==-2} {
      if {$vgl<$val} {
        set ret 1
      } elseif {$vgl>$val} {
        set ret 0
      }
    } elseif {$ret==0} {
      if {$vgl<$val} {return -1}
    } elseif {$ret==1} {
      if {$vgl>$val} {return -1}
    }
    set vgl $val
  }
  if {$ret==-3} {set ret -1}
  if {$ret==-2} {set ret 1}
  return $ret
}

proc cardRow {cardList} {
  # return iff the cardList is an ascending (1), descending (0) or no (-1)
  # row of cards. 1 is checked before 0. A row is a special sequence where
  # each card is a direct successor of the previous one (8,9,10,...)
  set ret -3
  foreach c $cardList {
    set val [cardValue [list $c]]
    if {$ret==-3} {
      set ret -2
    } elseif {$ret==-2} {
      if {[expr $val-$vgl]==1} {
        set ret 1
      } elseif {[expr $val-$vgl]==-1} {
        set ret 0
      } else {
        return -1
      }
    } elseif {$ret==0} {
      if {[expr $val-$vgl]!=-1} {return -1}
    } elseif {$ret==1} {
      if {[expr $val-$vgl]!=1} {return -1}
    }
    set vgl $val
  }
  if {$ret==-3} {set ret -1}
  if {$ret==-2} {set ret 1}
  return $ret
}

proc cardExamine {cardList} {
  # this proc returns info about a set of cards
  # return value is a list of tags:
  #   OPEN	all cards lie open
  #   FLIPPED	all cards show their backs
  #   UPSEQ	the list is an ascending sequence
  #   DOWNSEQ	the list is a descending sequence
  #   UPROW	the sequence is an ascending row
  #   DOWNROW	the sequence is a descending row
  #   COLOR	all cards are from the same color
  #   ALTCOL	the cards follow in alternate colors (red black red...)
  #   PICTURE	all cards have the same picture (karo..kreuz)
  #   VALUE	all cards have the same value
  set ret {}
  set fl [cardFlipped $cardList]
  if {$fl==0} {
    lappend ret OPEN
  } elseif {$fl==1} {
    lappend ret FLIPPED
  }
  if {[cardColor $cardList]!=-1} {lappend ret COLOR}
  if {[cardAltColors $cardList]} {lappend ret ALTCOL}
  if {[cardPicture $cardList]!=-1} {lappend ret PICTURE}
  if {[cardValue $cardList]!=-1} {lappend ret VALUE}
  set cs [cardSequence $cardList]
  if {$cs==0} {
    lappend ret DOWNSEQ
    if {[cardRow $cardList]==0} {lappend ret DOWNROW}
  } elseif {$cs==1} {
    lappend ret UPSEQ
    if {[cardRow $cardList]==1} {lappend ret UPROW}
  }
  return $ret
}

proc cardsHaveFlags {cardList flagList} {
  # return if the card list has the given flags
  set flags [cardExamine $cardList]
  foreach f $flagList {
    if {[lsearch -exact $flags $f]==-1} {return 0}
  }
  return 1
}

proc cardReadImages {dir imageListName backListName} {
  upvar $backListName bl
  upvar $imageListName il

  # read images and store their names
  set il {}
  for {set i 0} {$i<52} {incr i} {
    image create photo c$i -file $dir/$i.gif
    lappend il c$i
  }

  # read backs and store their names
  set bl {}
  foreach n [glob $dir/b*.gif] {
    set iname [file rootname [file tail $n]]
    image create photo $iname -file $n
    lappend bl $iname
  }
}
