set GAME(short) spider
set GAME(name) "Spider Solitaire"
set GAME(comment) "Version 2003.04.20
Copr (c) 1999-2002 Richard Breuer
richard.breuer@gmx.de / www.ricki.de"
set GAME(rules) "SPIDER SOLITAIRE

1. Foundations. (Top of layout)
    A. The foundations are not built on to directly.
    B. The goal is to create eight suites of 13 cards of the same suit from King to Ace.
    C. Complete these piles to win the game.

2. Tableau. (Below foundations)
    A. Build down regardless of suit. However, it is best to try and build in suit if possible.
    B. Any face up card in the tableau may be played to another pile as long as all cards above it are built in sequence and in suit.
    C. Nothing can be placed on an Ace.
    D. As face down cards become exposed, they can be flipped over and are now available for play.
    E. Empty spaces can be filled with any card.
    F. Cards are dealt to these piles from the stock.

3. Stock. (Bottom of layout)
    A. After all possible moves are made, one card is dealt out to each tableau pile to create more possibilities.

\[From www.solitaire-games.com\]"

# +-+     +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+
# |0|     |1| |2| |3| |4| |5| |6| |7| |8|
# +-+     +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+
#
# +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+
# |9| |0| |1| |2| |3| |4| |5| |6| |7| |8|
# +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+

array set DATA {
   STACKS	19

   0,x  50	 0,y  50	 0,xo	0	 0,yo	0

   1,x 210	 1,y  50	 1,xo	0	 1,yo	0
   2,x 290	 2,y  50	 2,xo	0	 2,yo	0
   3,x 370	 3,y  50	 3,xo	0	 3,yo	0
   4,x 450	 4,y  50 	 4,xo	0	 4,yo	0
   5,x 530	 5,y  50 	 5,xo	0	 5,yo	0
   6,x 610	 6,y  50 	 6,xo	0	 6,yo	0
   7,x 690	 7,y  50 	 7,xo	0	 7,yo	0
   8,x 770	 8,y  50 	 8,xo	0	 8,yo	0

   9,x  50	 9,y 200 	 9,xo	0	 9,yo	29
  10,x 130	10,y 200	10,xo	0	10,yo	29
  11,x 210	11,y 200	11,xo	0	11,yo	29
  12,x 290	12,y 200	12,xo	0	12,yo	29
  13,x 370	13,y 200	13,xo	0	13,yo	29
  14,x 450	14,y 200	14,xo	0	14,yo	29
  15,x 530	15,y 200	15,xo	0	15,yo	29
  16,x 610	16,y 200	16,xo	0	16,yo	29
  17,x 690	17,y 200	17,xo	0	17,yo	29
  18,x 770	18,y 200	18,xo	0	18,yo	29
}

proc doMove {fromStack toStack count} {
  set redo "stackMove $fromStack $toStack $count"
  set undo "stackMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doRevMove {fromStack toStack count} {
  set redo "stackRevMove $fromStack $toStack $count"
  set undo "stackRevMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doFlipMove {fromStack toStack count} {
  set redo "stackMove $fromStack $toStack $count
            stackFlip $fromStack 1"
  set undo "stackFlip $fromStack 1
            stackMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doRevFlipMove {fromStack toStack count} {
  set redo "stackRevMove $fromStack $toStack $count
            stackFlip $fromStack 1"
  set undo "stackFlip $fromStack 1
            stackRevMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc normalMove {fromStack toStack count {ideal 0}} {
  # move from stack 9..18 to 9..18; ideal tries an ideal, mathching move

  # only rows of matching cards may be moved
  set cl [stackGet $fromStack $count]
  if {![cardsHaveFlags $cl {OPEN PICTURE UPROW}]} {return 0}

  # the uppermost card must match the target stack
  if {![stackIsEmpty $toStack]} {
    set cl [lindex $cl end]
    if {$ideal && [cardPicture [stackGet $toStack 1]]!=[cardPicture $cl]} {return 0}
    if {[expr [cardValue [stackGet $toStack 1]]-[cardValue $cl]]!=1} {return 0}
  }

  # eventually flip
  set cl [stackGet $fromStack [expr $count+1]]
  set cl [lindex $cl end]
  if {[cardFlipped $cl]==1} {
    doRevFlipMove $fromStack $toStack $count
  } else {
    doRevMove $fromStack $toStack $count
  }

  winner
  return 1
}

proc finalMove {fromStack toStack count} {
  # move from stack 9..18 to 1..8

  # the target stack must be empty
  if {![stackIsEmpty $toStack]} {return 0}

  # only complete rows of matching cards may be moved
  set cl [stackGet $fromStack $count]
  if {$count!=13 || ![cardsHaveFlags $cl {OPEN PICTURE UPROW}]} {return 0}

  # eventually flip
  set cl [stackGet $fromStack [expr $count+1]]
  set cl [lindex $cl end]
  if {[cardFlipped $cl]==1} {
    doRevFlipMove $fromStack $toStack $count
  } else {
    doRevMove $fromStack $toStack $count
  }

  winner
  return 1
}

# THE FOLLOWING MUST BE IMPLEMENTED!!!

proc move {fromStack toStack {count 1}} {
  # never accept cards from the upper stacks or to stack 0
  if {$fromStack<9} {return 0}
  if {$toStack==0} {return 0}

  # check for three possible moves
  if {$toStack>8} {
    if {[normalMove $fromStack $toStack $count]} {autoPlay; return 1}
  } elseif {$toStack>0} {
    if {[finalMove $fromStack $toStack $count]} {autoPlay; return 1}
  }

  return 0
}

proc specialAction {stack {count 1} {type RIGHT}} {

  if {$stack==0} {
    # You may only deal if no stack is free
    for {set s 9} {$s<19} {incr s} {
      if {[stackIsEmpty $s]} {
        message "You cannot deal until no stacks are empty!"
        return 0
      }
    }

    # deal
    set redo {foreach s {9 10 11 12 13 14 15 16 17 18} {
                stackMoveFlipped 0 $s 1
              }}
    set undo {foreach s {9 10 11 12 13 14 15 16 17 18} {
                stackMoveFlipped $s 0 1
              }}
    moveAdd $redo $undo
    return 1
  } elseif {9<=$stack} {
    # get the longest streak of movable cards
    set cl [stackGetFlagged $stack {OPEN PICTURE UPROW}]
    set cll [llength $cl]

    # try to place the stack, shorten it and try again...
    while {$cll>0} {
      # first try the ideal moves
      for {set s 9} {$s<19} {incr s} {
        if {$s!=$stack && [normalMove $stack $s $cll 1]} {return 1}
      }
      # then try the normal moves
      for {set s 9} {$s<19} {incr s} {
        if {$s!=$stack && [normalMove $stack $s $cll 0]} {return 1}
      }
      incr cll -1
    }
  }
  return 0
}

proc autoPlay {{explicitCall 0}} {}

proc finish {} {
  # return 1 for winner, -1 for loser, and 0 for game on
  for {set s 9} {$s<=18} {incr s} {
    if {![stackIsEmpty $s]} {return 0}
  }
  return 1
}

proc initGame {{startSeed ""}} {
  global CARDS OPTIONS GAME widget

  set GAME(start) [clock seconds]

  # OPTIONS
  # -------
  # autoplay		Choose the autoplay feature

  set OPTIONS(autoplay) 0

  wm geometry $widget(topwindow) 830x700

  stackRemoveCards all all

  set CARDS {}
  for {set i 0} {$i<104} {incr i} {lappend CARDS [expr -$i-1]}

  set GAME(seed) $startSeed
  if {$GAME(seed)==""} {
    set GAME(seed) [expr [clock seconds]%65536]
  }

  mix $GAME(seed)

  set p 0
  for {set s 0} {$s<10} {incr s} {
    if {[expr $s%3]==0} {
      stackAddCards [expr $s+9] [lrange $CARDS $p [expr $p+4]] 0
      incr p 5
      stackAddCards [expr $s+9] [expr -[lindex $CARDS $p]-1] 0
      incr p 1
    } else {
      stackAddCards [expr $s+9] [lrange $CARDS $p [expr $p+3]] 0
      incr p 4
      stackAddCards [expr $s+9] [expr -[lindex $CARDS $p]-1] 0
      incr p 1
    }
  }
  stackAddCards 0 [lrange $CARDS $p end] 0

  stackRefresh
  moveInit

  autoPlay
}
