set GAME(short) towers
set GAME(name) "Seahaven Towers Solitaire"
set GAME(comment) "Version 2003.04.20
Copr (c) 1999-2002 Richard Breuer
richard.breuer@gmx.de / www.ricki.de"
set GAME(rules) "SEAHAVEN TOWERS SOLITAIRE

1. Foundations (2 times 2). (Top Left and Right of layout)
    A. Build up in suit from Ace to King.
    B. Complete these piles to win the game.

2. Tableau (10). (Below foundations)
    A. Build down in suit.
    B. The topmost card is available for play on the foundations.
    C. Any built unit of cards can be moved to another tableau pile as long as there are enough free reserves to hold them.
    D. Empty spaces can be filled by a King or a pile whose bottommost card is a King.

3. Reserves (4). (Between foundations)
    A. These are the 'cells'.
    B. Cards in these piles can be played on the foundations and the tableau.
    C. Each reserve can hold one card.
    
\[From www.solitaire-games.com\]"

# +-+ +-+     +-+ +-+ +-+ +-+     +-+ +-+
# |0| |1|     |4| |5| |6| |7|     |2| |3|
# +-+ +-+     +-+ +-+ +-+ +-+     +-+ +-+
#
# +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+
# |8| |9| |0| |1| |2| |3| |4| |5| |6| |7|
# +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+

array set DATA {
   STACKS	18

   0,x  50	 0,y  50	 0,xo	0	 0,yo	0
   1,x 130	 1,y  50	 1,xo	0	 1,yo	0
   2,x 690	 2,y  50	 2,xo	0	 2,yo	0
   3,x 770	 3,y  50	 3,xo	0	 3,yo	0

   4,x 290	 4,y  50 	 4,xo	0	 4,yo	0
   5,x 370	 5,y  50 	 5,xo	0	 5,yo	0
   6,x 450	 6,y  50 	 6,xo	0	 6,yo	0
   7,x 530	 7,y  50 	 7,xo	0	 7,yo	0

   8,x  50	 8,y 200 	 8,xo	0	 8,yo	29
   9,x 130	 9,y 200	 9,xo	0	 9,yo	29
  10,x 210	10,y 200	10,xo	0	10,yo	29
  11,x 290	11,y 200	11,xo	0	11,yo	29
  12,x 370	12,y 200	12,xo	0	12,yo	29
  13,x 450	13,y 200	13,xo	0	13,yo	29
  14,x 530	14,y 200	14,xo	0	14,yo	29
  15,x 610	15,y 200	15,xo	0	15,yo	29
  16,x 690	16,y 200	16,xo	0	16,yo	29
  17,x 770	17,y 200	17,xo	0	17,yo	29
}

proc emptyStacks {} {
  return [expr [stackIsEmpty 4]+[stackIsEmpty 5]+[stackIsEmpty 6]+[stackIsEmpty 7]]
}

proc doMove {fromStack toStack count} {
  set redo "stackMove $fromStack $toStack $count"
  set undo "stackMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doRevMove {fromStack toStack count} {
  set redo "stackRevMove $fromStack $toStack $count"
  set undo "stackRevMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc normalMove {fromStack toStack count} {
  if {$count>[expr 1+[emptyStacks]]} {return 0}
  set cl [stackGet $fromStack $count]
  if {![cardsHaveFlags $cl {OPEN UPROW PICTURE}]} {return 0}
  if {[stackIsEmpty $toStack]} {
    if {[cardValue [lindex $cl end]]!=12} {return 0}
  } else {
    if {[cardValue [lindex $cl end]]!=[expr [cardValue [stackGet $toStack 1]]-1]} {return 0}
    if {[cardPicture [lindex $cl 0]]!=[cardPicture [stackGet $toStack 1]]} {return 0}
  }
  doRevMove $fromStack $toStack $count
  winner
  return 1
}

proc upperMove {fromStack toStack count} {
  if {$count>[emptyStacks]} {return 0}
  # place up to 4 cards
  set s $toStack
  while {$count>0} {
    if {[stackIsEmpty $s]} {
      doMove $fromStack $s 1
      incr count -1
    }
    set s [expr (($s-4+1)%4)+4]
  }
  winner
  return 1
}

proc finalMove {fromStack toStack count} {
  set cl [stackGet $fromStack $count]
  if {![cardsHaveFlags $cl {OPEN UPROW PICTURE}]} {return 0}
  if {[stackIsEmpty $toStack]} {
    if {[cardValue [lindex $cl 0]]!=0} {return 0}
  } else {
    if {[cardValue [lindex $cl 0]]!=[expr 1+[cardValue [stackGet $toStack 1]]]} {return 0}
    if {[cardPicture [lindex $cl 0]]!=[cardPicture [stackGet $toStack 1]]} {return 0}
  }
  doMove $fromStack $toStack $count
  winner
  return 1
}

# THE FOLLOWING MUST BE IMPLEMENTED!!!

proc move {fromStack toStack {count 1}} {
  # never accept cards from the towers
  if {$fromStack<4} {return 0}

  # check for three possible moves
  if {$toStack>7} {
    if {[normalMove $fromStack $toStack $count]} {autoPlay; return 1}
  } elseif {$toStack>3} {
    if {[upperMove $fromStack $toStack $count]} {autoPlay; return 1}
  } else {
    if {[finalMove $fromStack $toStack $count]} {autoPlay; return 1}
  }

  return 0
}

proc specialAction {stack {count 1} {type RIGHT}} {
  return 0
}

proc autoPlay {{explicitCall 0}} {
  global OPTIONS

  if {$OPTIONS(autoplay)==0} return
  if {$OPTIONS(autoplay)==1 && !$explicitCall} return

  block

  # check for auto moves until the game is over or none is left
  set ex 0
  while {!$ex} {
    # check all final and working stacks for a move
    set ex 1
    for {set t 0} {$t<4} {incr t} {
      for {set s 4} {$s<18} {incr s} {
        if {[finalMove $s $t 1]} {
          while {[finalMove $s $t 1]} {}
          if {[winner]!=0} {unblock; return}
          set ex 0
        }
      }
    }
  }
  unblock
}

proc finish {} {
  # return 1 for winner, -1 for loser, and 0 for game on
  for {set s 4} {$s<=17} {incr s} {
    if {![stackIsEmpty $s]} {return 0}
  }
  return 1
}

proc initGame {{startSeed ""}} {
  global CARDS OPTIONS GAME widget

  set GAME(start) [clock seconds]

  # OPTIONS
  # -------
  # autoplay		Choose the autoplay feature

  set OPTIONS(autoplay) 2

  wm geometry $widget(topwindow) 830x650

  stackRemoveCards all all

  set CARDS {}
  for {set i 0} {$i<52} {incr i} {lappend CARDS [expr $i]}

  set GAME(seed) $startSeed
  if {$GAME(seed)==""} {
    set GAME(seed) [expr [clock seconds]%65536]
  }

  mix $GAME(seed)

  for {set s 0} {$s<10} {incr s} {
    stackAddCards [expr $s+8] [lrange $CARDS [expr 5*$s] [expr 5*$s+4]] 0
  }
  stackAddCards 5 [lindex $CARDS 50] 0
  stackAddCards 6 [lindex $CARDS 51] 0

  stackRefresh
  moveInit

  autoPlay
}
